/**
 * Copyright (c) 2013-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import java.io.IOException;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.TimeUnit;
import java.util.logging.Logger;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;

import org.objectweb.fractal.fraclet.annotations.Component;
import org.objectweb.fractal.fraclet.annotations.Interface;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.types.Step;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.api.communication.sharedarea.SharedAreaService;
import org.ow2.petals.microkernel.api.communication.sharedarea.exception.InvalidPropertyException;
import org.ow2.petals.microkernel.api.communication.sharedarea.exception.MissingPropertyException;
import org.ow2.petals.microkernel.api.communication.sharedarea.exception.SharedAreaException;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.implementation.exception.MutableImplException;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.microkernel.registry.overlay.api.RegistryOverlayAPI;
import org.ow2.petals.microkernel.registry.overlay.exception.AuthenticationException;
import org.ow2.petals.microkernel.registry.overlay.exception.InvalidConfigurationException;
import org.ow2.petals.registry_overlay.api.Constants;
import org.ow2.petals.registry_overlay.api.DataModelApi;
import org.ow2.petals.registry_overlay.api.PetalsRegistryPortableFactory;
import org.ow2.petals.registry_overlay.api.StoredRegistryEndpoint;
import org.ow2.petals.registry_overlay.api.StoredRegistryEndpoint.Type;
import org.ow2.petals.registry_overlay.client.configuration.ConfigurationBuilder;
import org.ow2.petals.registry_overlay.client.configuration.ConfigurationException;
import org.ow2.petals.registry_overlay.client.configuration.generated.Configuration;
import org.ow2.petals.registry_overlay.client.configuration.generated.OverlayMember;
import org.ow2.petals.registry_overlay.client.configuration.generated.OverlayMembers;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.Topology;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.ebmwebsourcing.easycommons.xml.DocumentBuilders;
import com.ebmwebsourcing.easycommons.xml.XMLHelper;
import com.google.common.base.Function;
import com.google.common.base.Optional;
import com.google.common.collect.Collections2;
import com.hazelcast.client.HazelcastClient;
import com.hazelcast.client.config.ClientConfig;
import com.hazelcast.client.config.ClientNetworkConfig;
import com.hazelcast.config.NearCacheConfig;
import com.hazelcast.config.SerializationConfig;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.core.IMap;
import com.hazelcast.map.EntryProcessor;
import com.hazelcast.query.Predicate;
import com.hazelcast.query.Predicates;
import com.hazelcast.query.SqlPredicate;
import com.hazelcast.query.TruePredicate;

/**
 * The registry implementation based on the Petals Registry Overlay.
 * 
 * Note: some of these methods (see e.g., {@link #getEndpoints(EndpointType)}) return lazy collection (aka live view)
 * but they never relies on hazelcast: once the collection is returned, no hazelcast requests are done!
 * 
 * TODO make the live-view not recompute their value everytime they are called!
 * 
 * TODO should we do something specific such a throwing errors if the shared area is used after we left the topology?!
 * 
 * @author Christophe DENEUX - Linagora
 * @author Victor Noel
 */

@Component(provides = @Interface(name = SharedAreaService.FRACTAL_SRV_ITF_NAME, signature = SharedAreaService.class))
public class RegistryOverlayImpl implements RegistryOverlayAPI {

    /**
     * Prefix of property names
     */
    private static final String PROPERTY_NAME_PREFIX = RegistryOverlayImpl.class.getName() + ".";
    
    /**
     * Prefix of property names configuring map cache
     */
    private static final String PROPERTY_NAME_MAP_CACHE_PREFIX = PROPERTY_NAME_PREFIX + "map-cache.";

    /**
     * Name of the property defining the max size of map cache, in endpoint number
     */
    public static final String PROPERTY_NAME_MAP_CACHE_MAX_SIZE = PROPERTY_NAME_MAP_CACHE_PREFIX + "max-size";

    /**
     * Name of the property associated to the group name
     */
    private static final String GROUP_NAME_PROPERTY_NAME = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.GROUP_NAME;
    
    /**
     * Name of the property associated to the group password
     */
    private static final String GROUP_PASSWORD_PROPERTY_NAME = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.GROUP_PASSWORD;

    /**
     * Name pattern of the property associated to an overlay member host name
     */
    private static final String MEMBER_HOST_PROPERTY_NAME_PATTERN = PROPERTY_NAME_PREFIX
            + RegistryOverlayAPI.RegistryItem.ADDRESS + ".<x>";

    /**
     * Default value of the property {@link #PROPERTY_NAME_MAP_CACHE_MAX_SIZE}
     */
    public static final int DEFAULT_MAP_CACHE_MAX_SIZE = 500;

    /**
     * The map containing endpoints.
     */
    private IMap<String, StoredRegistryEndpoint> endpointsMap;

    /**
     * <p>
     * The map containing endpoint description.
     * </p>
     * <p>
     * To avoid classloader problem, the WSDL is stored as {@link String}.
     * </p>
     */
    private IMap<String, String> endpointDescriptionsMap;

    /**
     * <p>
     * The map containing dynamic topologies.
     * </p>
     * <p>
     * Note: Type is {@link IMap} to be able to register a listener on updated entries.
     * </p>
     */
    private IMap<String, Topology> topologiesMap;

    /**
     * Identifier of the listener on the topology entries
     */
    private String dynamicTopologyListener;

    /**
     * Configuration service fractal component
     */
    @Requires(name = SharedAreaService.CONFIGURATION_FRACTAL_ITF_NAME)
    protected ConfigurationService configurationService;

    /**
     * Logging utilities
     */
    private final LoggingUtil log = new LoggingUtil(Logger.getLogger(COMPONENT_LOGGER_NAME));

    /**
     * The topology name associated to the current domain in the registry
     */
    private String topologyName;

    private ContainerConfiguration containerConfiguration;

    /**
     * The Hazelcast client
     */
    private HazelcastInstance hzlClient;

    private Configuration registryOverlayConfiguration = null;

    @Override
    public Collection<PetalsServiceEndpoint> getEndpoints(final EndpointType type) throws SharedAreaException {
        try {

            Predicate<?, ?> p = TruePredicate.INSTANCE;

            if (type != null) {
                p = matchType(type);
            }

            // values() returns a copy of the values
            return Collections2.transform(this.endpointsMap.values(p), Stored2Petals.INSTANCE);
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    /**
     * TODO See {@link ExternalEndpointImpl} for PETALSESBCONT-393
     */
    @Override
    public PetalsServiceEndpoint registerExternalEndpoint(final ServiceEndpoint endpoint, final String componentName)
            throws SharedAreaException {
        // we pack it so that it can be serialized in the shared area!
        final StoredRegistryEndpoint storedEndpoint = toStoredEndpoint(Type.EXTERNAL, endpoint,
                new Location(this.containerConfiguration.getName(), componentName));
        if (registerEndpoint(storedEndpoint, null)) {
            return new ExternalEndpointImpl(storedEndpoint);
        } else {
            return null;
        }
    }

    @Override
    public PetalsServiceEndpoint registerInternalEndpoint(final String endpointName, final QName serviceName,
            final QName[] interfaces, final String componentName, final Document description)
            throws SharedAreaException {


        final StoredRegistryEndpoint storedEndpoint = toStoredEndpoint(Type.INTERNAL, endpointName, serviceName,
                interfaces, new Location(this.containerConfiguration.getName(), componentName));
        if (registerEndpoint(storedEndpoint, description)) {
            return new InternalEndpointImpl(storedEndpoint);
        } else {
            return null;
        }
    }

    private boolean registerEndpoint(final StoredRegistryEndpoint serviceEndpoint, final Document description)
            throws SharedAreaException {

        if (serviceEndpoint == null) {
            throw new SharedAreaException("Can not register a null service endpoint");
        }

        final String key = getEndpointKey(serviceEndpoint);

        try {
            // the lock is here only to handle the case where two different container/component try to register the
            // same endpoint. It is also used when removing endpoints to avoid interleaving especially when we want to
            // get the old value. It covers both endpoints and description maps.
            final int lockWaitTime = this.containerConfiguration.getTopologyLocksMaxWaitTime();
            if (this.endpointsMap.tryLock(key, lockWaitTime, TimeUnit.SECONDS)) {
                try {
                    // We register the endpoint
                    final StoredRegistryEndpoint previous = this.endpointsMap.putIfAbsent(key, serviceEndpoint);
                    if (previous != null) {
                        final String localContainerName = this.containerConfiguration.getName();
                        if (previous.isInRestoration() && previous.getContainer().equals(localContainerName)) {
                            // in that case we can simply replace it!
                            // note: we don't verify that the component is the same, the only important point is that a
                            // previously activated endpoint is reactivated by the same container, all the rest is out
                            // of scope of this class.
                            this.endpointsMap.set(key, serviceEndpoint);
                        } else {
                            // in that case, we can't add it
                            return false;
                        }
                    }

                    // For performance reasons, the description is not saved as part of the service endpoint. It is
                    // retrieved only if needed through ServiceEndpoint.getDescription(). See #retrieveDescription()
                    // used by RegistryOverlayEndpoint.
                    if (description != null) {
                        try {
                            // we omit declaration, it will save a bit of space...
                            this.endpointDescriptionsMap.set(key, XMLHelper.createStringFromDOMNode(description, true));
                        } catch (final TransformerException e) {
                            throw new SharedAreaException("Can't serialize the description into the registry", e);
                        }
                    }

                    return true;
                } finally {
                    this.endpointsMap.unlock(key);
                }
            } else {
                throw new SharedAreaException(String
                        .format("The lock on the endpoint %s is already acquired (waited %s).", key, lockWaitTime));
            }
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        } catch (final InterruptedException e) {
            throw new SharedAreaException("Locking on endpoint " + key + " interrupted");
        }
    }

    private static String getEndpointKey(final StoredRegistryEndpoint serviceEndpoint) {
        return getEndpointKey(serviceEndpoint.getServiceName(), serviceEndpoint.getEndpointName(),
                serviceEndpoint.getType());
    }

    private static String getEndpointKey(final QName serviceName, final String endpointName, final Type type) {
        return getEndpointKey(serviceName.toString(), endpointName, type);
    }

    private static String getEndpointKey(final String serviceName, final String endpointName, final Type type) {
        return (type == Type.INTERNAL ? "I" : "E") + "<" + serviceName + ">" + endpointName;
    }

    @Override
    public PetalsServiceEndpoint deregisterEndpoint(final QName serviceName, final String endpointName,
            final String componentName, final EndpointType type) throws SharedAreaException {
        final String key = getEndpointKey(serviceName, endpointName, toStoredType(type));
        try {
            final int lockWaitTime = this.containerConfiguration.getTopologyLocksMaxWaitTime();
            // see registerEndpoint
            if (this.endpointsMap.tryLock(key, lockWaitTime, TimeUnit.SECONDS)) {
                try {
                    final StoredRegistryEndpoint stored = this.endpointsMap.get(key);

                    // TODO use an EntryProcessor, it will simplify locking and limit the number of actions...
                    if (stored != null) {
                        final String containerName = this.containerConfiguration.getName();
                        if (!stored.getComponent().equals(componentName)
                                || !stored.getContainer().equals(containerName)) {
                            this.log.error("Component " + componentName + " tried to deactivate endpoint " + key
                                    + " owned by component " + stored.getComponent() + " on container "
                                    + stored.getContainer());
                            return null;
                        }

                        // use delete because we don't need the value
                        this.endpointsMap.delete(key);
                        this.endpointDescriptionsMap.delete(key);

                        return Stored2Petals.INSTANCE.apply(stored);
                    } else {
                        return null;
                    }
                } finally {
                    this.endpointsMap.unlock(key);
                }
            } else {
                throw new SharedAreaException(String
                        .format("The lock on the endpoint %s is already acquired (waited %s).", key, lockWaitTime));
            }
        } catch (final InterruptedException e) {
            throw new SharedAreaException("Locking on endpoint " + key + " interrupted");
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getEndpointsForInterface(final QName interfaceName,
            final EndpointType type) throws SharedAreaException {

        try {
            Predicate<?, ?> p = matchInterface(interfaceName);

            if (type != null) {
                p = Predicates.and(p, matchType(type));
            }
            
            return Collections2.transform(this.endpointsMap.values(p), Stored2Petals.INSTANCE);
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public Collection<PetalsServiceEndpoint> getEndpointsForService(final QName serviceName, final EndpointType type)
            throws SharedAreaException {

        try {
            Predicate<?, ?> p = matchService(serviceName);

            if (type != null) {
                p = Predicates.and(p, matchType(type));
            }

            return Collections2.transform(this.endpointsMap.values(p), Stored2Petals.INSTANCE);
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public PetalsServiceEndpoint getEndpoint(final QName service, final String endpointName, final EndpointType type)
            throws SharedAreaException {

        assert service != null;
        assert endpointName != null;

        if (this.log.isDebugEnabled()) {
            this.log.start("with params serviceName = " + service.toString() + ", endpointName = " + endpointName);
        }

        try {
            final StoredRegistryEndpoint stored = this.endpointsMap
                    .get(getEndpointKey(service, endpointName, toStoredType(type)));
            return stored != null ? Stored2Petals.INSTANCE.apply(stored) : null;
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * TODO Improve using an {@link EntryProcessor} maybe?
     */
    private Collection<SavedEndpoint> removeLocalEndpointsWithDocument() throws SharedAreaException {
        final Collection<StoredRegistryEndpoint> localEndpoints = getLocalEndpoints(false);
        // let's build a collection
        final Collection<SavedEndpoint> endpoints = new ArrayList<>(localEndpoints.size());
        for (final StoredRegistryEndpoint endpoint : localEndpoints) {
            final String key = getEndpointKey(endpoint);
            try {
                // TODO or log in case of problems?
                final int lockWaitTime = this.containerConfiguration.getTopologyLocksMaxWaitTime();
                // see registerEndpoint
                if (this.endpointsMap.tryLock(key, lockWaitTime, TimeUnit.SECONDS)) {
                    try {
                        final Document description;
                        if (endpoint.getType().equals(StoredRegistryEndpoint.Type.INTERNAL)) {
                            description = toDocument(this.endpointDescriptionsMap.remove(key));
                        } else {
                            description = null;
                        }
                        this.endpointsMap.delete(key);
                        endpoints.add(new SavedEndpoint(Stored2Petals.INSTANCE.apply(endpoint), description));
                    } finally {
                        this.endpointsMap.unlock(key);
                    }
                } else {
                    throw new SharedAreaException(String.format(
                            "The lock on the endpoint %s is already acquired (waited %s).", key, lockWaitTime));
                }
            } catch (final InterruptedException e) {
                throw new SharedAreaException("Locking on endpoint " + key + " interrupted");
            }
        }
        return endpoints;
    }

    private Collection<StoredRegistryEndpoint> getLocalEndpoints(final boolean onlyInRestoration) {

        Predicate<?, ?> cont = Predicates.equal(DataModelApi.ENDPOINT_CONTAINER_INDEX,
                this.containerConfiguration.getName());

        if (onlyInRestoration) {
            cont = Predicates.and(cont, Predicates.equal(DataModelApi.ENDPOINT_RESTAURATION_INDEX, true));
        }

        return this.endpointsMap.values(cont);
    }

    @Lifecycle(step = Step.START)
    public void start() throws SharedAreaException {
        this.log.start();

        try {
            final String domainName = this.configurationService.getDomainConfiguration().getName();
            if (domainName == null) {
                throw new InvalidConfigurationException("The domain name is missing");
            }
            this.topologyName = domainName;

            this.containerConfiguration = this.configurationService.getContainerConfiguration();

            final List<Object> extraParameters = this.configurationService.getDomainConfiguration()
                    .getExtraParameters();
            if (extraParameters != null) {
                try {
                    for (final Object extraParameter : extraParameters) {
                        final Configuration cfg = ConfigurationBuilder
                                .createConfiguration((org.w3c.dom.Node) extraParameter);
                        if (cfg != null) {
                            this.registryOverlayConfiguration = cfg;
                            break;
                        }
                    }
                } catch (final ConfigurationException e) {
                    throw new InvalidConfigurationException(e);
                }
                if (this.registryOverlayConfiguration == null) {
                    throw new InvalidConfigurationException("Missing configuration");
                }
            } else {
                throw new InvalidConfigurationException("Missing configuration");
            }

            if (this.registryOverlayConfiguration.getGroupName() == null
                    || this.registryOverlayConfiguration.getGroupName().isEmpty()) {
                throw new InvalidConfigurationException("The group name is missing");
            }
            if (this.registryOverlayConfiguration.getGroupPassword() == null
                    || this.registryOverlayConfiguration.getGroupPassword().isEmpty()) {
                throw new InvalidConfigurationException("The group password is missing");
            }
            if (this.registryOverlayConfiguration.getOverlayMembers() == null
                    || this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember() == null
                    || this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember().isEmpty()) {
                throw new InvalidConfigurationException("No overlay member declared");
            }

            if (this.log.isConfigEnabled()) {
                this.log.config("Registry name: " + this.topologyName);
                this.log.config("Group name: " + this.registryOverlayConfiguration.getGroupName());
                this.log.config("Group pwd: ****");
            }

            // Connect to the Hazelcast cluster
            final ClientConfig cltCfg = new ClientConfig();

            cltCfg.getGroupConfig().setName(this.registryOverlayConfiguration.getGroupName());
            cltCfg.getGroupConfig().setPassword(this.registryOverlayConfiguration.getGroupPassword());
            final ClientNetworkConfig networkConfig = new ClientNetworkConfig();
            cltCfg.setNetworkConfig(networkConfig);
            for (final OverlayMember overlayMember : this.registryOverlayConfiguration.getOverlayMembers()
                    .getOverlayMember()) {
                final String host = overlayMember.getValue();
                final BigInteger port = overlayMember.getPort();
                if (host == null || host.isEmpty()) {
                    throw new InvalidConfigurationException("The hostname of an overlay member is missing");
                }
                final String address = host + ":" + (port == null ? Constants.DEFAULT_PORT : port.toString());
                if (this.log.isConfigEnabled()) {
                    this.log.config("Member: " + address);
                }
                networkConfig.addAddress(address);
            }

            // Enable near cache
            this.enableNearCache(cltCfg);

            // register the serialization factory
            cltCfg.setSerializationConfig(new SerializationConfig().addPortableFactoryClass(
                    PetalsRegistryPortableFactory.FACTORY_ID, PetalsRegistryPortableFactory.class));

            // We use the Petals ESB container classloader
            cltCfg.setClassLoader(Thread.currentThread().getContextClassLoader());

            try {
                this.hzlClient = HazelcastClient.newHazelcastClient(cltCfg);

                // Distributed objects about endpoints
                this.endpointsMap = DataModelApi.getEndpointsMap(this.hzlClient, this.topologyName);
                this.endpointDescriptionsMap = DataModelApi.getEndpointDescriptionsMap(this.hzlClient,
                        this.topologyName);

                this.topologiesMap = DataModelApi.getTopologiesMap(this.hzlClient);

            } catch (final RuntimeException e) {
                if (e.getCause() != null && e.getCause() instanceof com.hazelcast.client.AuthenticationException) {
                    throw new AuthenticationException();
                } else {
                    throw new SharedAreaException(e);
                }
            }
        } catch (final SharedAreaException e) {
            throw e;
        } catch (final Exception e) {
            throw new SharedAreaException(e);
        } finally {
            this.log.end();
        }
    }

    /**
     * TODO improve using {@link EntryProcessor}s maybe?
     */
    @Override
    public void restoreEndpoints(final Collection<SavedEndpoint> endpoints)
            throws SharedAreaException {
        this.log.start();

        try {
            // first we restore the saved endpoints
            for (final SavedEndpoint endpoint : endpoints) {
                final PetalsServiceEndpoint activated;
                if (endpoint.endpoint.getType() == EndpointType.EXTERNAL) {
                    activated = registerExternalEndpoint(endpoint.endpoint,
                            endpoint.endpoint.getLocation().getComponentName());
                } else if (endpoint.endpoint.getType() == EndpointType.INTERNAL) {
                    activated = registerInternalEndpoint(endpoint.endpoint.getEndpointName(),
                            endpoint.endpoint.getServiceName(), endpoint.endpoint.getInterfaces(),
                            endpoint.endpoint.getLocation().getComponentName(), endpoint.description);
                } else {
                    throw new SharedAreaException("Impossible case");
                }
                if (activated == null) {
                    this.log.warning("Endpoint not restored after joining topology "
                            + "(there is one already there from another container): " + endpoint.endpoint);
                }
            }

            // then we remove those that were not restored
            final Collection<PetalsServiceEndpoint> removed = removeLocalEndpointsInRestoration();
            for (final PetalsServiceEndpoint r : removed) {
                this.log.warning("Endpoint not restored after joining topology"
                        + " (it was there before but was not reactivated): " + r);
            }
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    private Collection<PetalsServiceEndpoint> removeLocalEndpointsInRestoration() throws SharedAreaException {
        final Collection<StoredRegistryEndpoint> localEndpoints = getLocalEndpoints(true);
        final Collection<PetalsServiceEndpoint> res = new ArrayList<>(localEndpoints.size());
        for (final StoredRegistryEndpoint endpoint : localEndpoints) {
            final String key = getEndpointKey(endpoint);
            try {
                // TODO or log in case of problems?
                final int lockWaitTime = this.containerConfiguration.getTopologyLocksMaxWaitTime();
                // see registerEndpoint
                if (this.endpointsMap.tryLock(key, lockWaitTime, TimeUnit.SECONDS)) {
                    try {
                        this.endpointDescriptionsMap.delete(key);
                        this.endpointsMap.delete(key);
                        res.add(Stored2Petals.INSTANCE.apply(endpoint));
                    } finally {
                        this.endpointsMap.unlock(key);
                    }
                } else {
                    throw new SharedAreaException(String.format(
                            "The lock on the endpoint %s is already acquired (waited %s).", key, lockWaitTime));
                }
            } catch (final InterruptedException e) {
                throw new SharedAreaException("Locking on endpoint " + key + " interrupted");
            }
        }
        return res;
    }

    @Override
    public void joinTopology(final boolean addMyself) throws SharedAreaException {

        if (this.dynamicTopologyListener != null) {
            throw new SharedAreaException("Can't load the topology if we already attached to one!");
        }

        final int lockWaitTime = this.containerConfiguration.getTopologyLocksMaxWaitTime();

        try {
            if (this.topologiesMap.tryLock(this.topologyName, lockWaitTime, TimeUnit.SECONDS)) {
                try {
                    Topology topology = this.topologiesMap.get(this.topologyName);
                    if (topology == null) {
                        // topology is not in the registry, let's use the local one to bootstrap the domain in the
                        // registry
                        topology = this.configurationService.getLocalTopology();
                        this.log.info("First startup of a Petals ESB node of the topology, the dynamic topology"
                                + " is initialized with the local topology definition.");
                        if (this.log.isDebugEnabled()) {
                            this.log.debug("Applied topology: " + topology);
                        }

                        // note: the topology in the ConfigurationService was already validated
                    } else {
                        // A topology exists, we use it
                        if (this.log.isDebugEnabled()) {
                            this.log.debug("Topology used from the registry: " + topology);
                        }

                        // if it's a dynamic topology and we are not in it, let's add ourselves!
                        if (addMyself) {
                            if (topology.getDomain().getMode() != DomainMode.DYNAMIC) {
                                throw new SharedAreaException("Can't add myself to a static topology!");
                            } else if (TopologyHelper.findContainer(this.containerConfiguration.getName(),
                                    topology) != null) {
                                throw new SharedAreaException(
                                        "Can't add myself to a dynamic topology I'm already attached to!");
                            } else {
                                final Container containerAttached = ConfigurationServiceUtils
                                        .toTopologyContainer(this.containerConfiguration);
                                topology.getDomain().getContainer().add(containerAttached);
                            }
                        }

                        // this will validate and persist the retrieved/updated topology
                        this.configurationService.updateTopology(topology, true);
                    }

                    // It is required to use IMap.put() instead of IMap.set() because IMap.set() breaks the contract
                    // of EntryListener firing an EntryEvent with a null oldValue. A non-null oldValue is required
                    // by DynamicTopologyListener.entryUpdated() to update the topology changes.
                    this.topologiesMap.put(this.topologyName, topology);
                    
                    // let's mark the endpoint has being in restoration
                    // TODO use an EntryProcessor instead (it will also improve locking)
                    for (final StoredRegistryEndpoint endpoint : getLocalEndpoints(false)) {
                        this.endpointsMap.set(getEndpointKey(endpoint), new StoredRegistryEndpoint(endpoint, true));
                    }

                    // and finally activate the topology listener for future events
                    // (before releasing the lock, so that nobody can do modification before I register it!)
                    this.dynamicTopologyListener = this.topologiesMap.addEntryListener(
                            new DynamicTopologyListener(this.configurationService, this.topologyName, this.log),
                            this.topologyName, true);
                } finally {
                    this.topologiesMap.unlock(this.topologyName);
                }

            } else {
                throw new SharedAreaException(
                        String.format("The lock on the topology %s is already acquired (waited %s).", this.topologyName,
                                lockWaitTime));
            }
        } catch (final org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException e) {
            throw new SharedAreaException("Invalid topology", e);
        } catch (final InterruptedException e) {
            throw new SharedAreaException("Locking was interrupted");
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    /**
     * Enable NearCache for all distributed map
     * 
     * @param cltCfg
     *            The Hazelcast client configuration
     */
    private final void enableNearCache(final ClientConfig cltCfg) {

        final Map<String, String> extraConfProperties = this.containerConfiguration.getExtraConfiguration();
        int maxSize;
        if (extraConfProperties != null) {
            final String maxSizeStr = extraConfProperties.get(PROPERTY_NAME_MAP_CACHE_MAX_SIZE);
            if (maxSizeStr != null && !maxSizeStr.trim().isEmpty()) {
                try {
                    maxSize = Integer.parseInt(maxSizeStr);
                    this.log.config("Max size of map caches: " + maxSize);
                } catch (final NumberFormatException e) {
                    this.log.warning(String.format(
                            "Invalid value for the max size of map cache: '%s'. Default value used: '%s'", maxSizeStr,
                            DEFAULT_MAP_CACHE_MAX_SIZE));
                    maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
                }
            } else {
                this.log.config("The max size of map cache is not defined. Default value used: "
                        + DEFAULT_MAP_CACHE_MAX_SIZE);
                maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
            }
        } else {
            this.log.config("No extra configuration defined for the registry overlay client implementation. Default values used");
            maxSize = DEFAULT_MAP_CACHE_MAX_SIZE;
        }

        final NearCacheConfig cltNearCacheEndpointsCfg = new NearCacheConfig();
        cltNearCacheEndpointsCfg.setName(DataModelApi.getEndpointsMapName(this.topologyName));
        cltNearCacheEndpointsCfg.setMaxSize(maxSize);
        cltNearCacheEndpointsCfg.setInvalidateOnChange(true);
        cltNearCacheEndpointsCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheEndpointsCfg);

        final NearCacheConfig cltNearCacheEndpointDescriptionsCfg = new NearCacheConfig();
        cltNearCacheEndpointDescriptionsCfg.setName(DataModelApi.getEndpointDescriptionsMapName(this.topologyName));
        cltNearCacheEndpointDescriptionsCfg.setMaxSize(maxSize);
        cltNearCacheEndpointDescriptionsCfg.setInvalidateOnChange(true);
        cltNearCacheEndpointDescriptionsCfg.setEvictionPolicy("LRU");
        cltCfg.addNearCacheConfig(cltNearCacheEndpointDescriptionsCfg);

        // Note: we don't want to use near cache with the topologies because they changes a lot (w.r.t. to the use we do
        // of it) and near cache is really best for read-only things. Moreover there is overhead for checking if a value
        // must be invalidated so it's really not the best.
    }

    @Override
    public Collection<SavedEndpoint> leaveTopology() throws SharedAreaException {

        if (this.dynamicTopologyListener == null) {
            throw new SharedAreaException("Can't leave a topology if we didn't attach to one before");
        }

        // first we stop the listener because it's not relevant anymore
        // And also so that our leaving is not persisted in case other events arrive
        final boolean removed = this.topologiesMap.removeEntryListener(this.dynamicTopologyListener);
        assert removed;
        this.dynamicTopologyListener = null;

        try {
            // then we remove ourselves
            if (this.configurationService.getDomainConfiguration()
                    .getMode() == DomainConfiguration.DomainMode.DYNAMIC) {
                final int lockWaitTime = this.containerConfiguration.getTopologyLocksMaxWaitTime();
                final String myName = this.containerConfiguration.getName();

                if (this.topologiesMap.tryLock(this.topologyName, lockWaitTime, TimeUnit.SECONDS)) {
                    try {
                        final Topology topology = this.topologiesMap.get(this.topologyName);
                        final Iterator<Container> itContainers = topology.getDomain().getContainer().iterator();
                        while (itContainers.hasNext()) {
                            if (myName.equals(itContainers.next().getName())) {
                                itContainers.remove();
                                break;
                            }
                        }
                        this.topologiesMap.put(this.topologyName, topology);
                    } finally {
                        this.topologiesMap.unlock(this.topologyName);
                    }

                    return removeLocalEndpointsWithDocument();
                } else {
                    throw new SharedAreaException(String
                            .format("The lock on the topology " + this.topologyName
                                    + " is already acquired (waited %s).", lockWaitTime));
                }
            } else {
                throw new SharedAreaException("Can't leave a static topology");
            }
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        } catch (final InterruptedException e) {
            throw new SharedAreaException("Locking on the topology " + this.topologyName + " interrupted");
        }
    }

    @Lifecycle(step = Step.STOP)
    public void stop() {
        this.log.start();

        if (this.hzlClient != null) {
            if (this.dynamicTopologyListener != null) {
                try {
                    this.topologiesMap.removeEntryListener(this.dynamicTopologyListener);
                    this.dynamicTopologyListener = null;
                } catch (final Exception e) {
                    this.log.warning("An error occured while removing the topology listener", e);
                }
            }

            try {
                HazelcastClient.shutdown(this.hzlClient);
                this.hzlClient = null;
            } catch (final Exception e) {
                this.log.warning("An error occured while stopping the registry client", e);
            }
        }

        this.log.end();
    }

    @Override
    public Properties getConfiguration(final String providedTopologyPassphrase) {
        final Properties result = new Properties();
        if (this.registryOverlayConfiguration != null) {
            final boolean isSecurityOk = this.configurationService
                    .isSecurityTopologyPassphraseOk(providedTopologyPassphrase);
            if (this.log.isDebugEnabled()) {
                if (isSecurityOk) {
                    this.log.debug("Security validated. Sensible information will be returned.");
                } else {
                    this.log.debug("Security not validated. Sensible information will NOT be returned.");
                }
            }

            result.setProperty(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.GROUP_NAME,
                    this.registryOverlayConfiguration.getGroupName());
            if (isSecurityOk) {
                result.setProperty(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.GROUP_PASSWORD,
                        this.registryOverlayConfiguration.getGroupPassword());
            }

            int addressNb = 0;
            for (final OverlayMember member : this.registryOverlayConfiguration.getOverlayMembers().getOverlayMember()) {
                final String host = member.getValue();
                final BigInteger port = member.getPort();
                if (host != null && !host.trim().isEmpty()) {
                    result.setProperty(
                            PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.ADDRESS + "." + addressNb++,
                            host + ":" + (port == null ? Constants.DEFAULT_PORT : port.toString()));
                }
            }
        }
        return result;
    }

    @Override
    public Document getTopologyPartConfiguration(final Properties properties) throws MutableImplException {
        
        if (properties != null) {
            
            final Configuration xmlConfiguration = new Configuration();
    
            // Group name
            final String groupName = properties.getProperty(GROUP_NAME_PROPERTY_NAME);
            if (groupName != null) {
                xmlConfiguration.setGroupName(groupName);
            } else {
                throw new MissingPropertyException(GROUP_NAME_PROPERTY_NAME);
            }

            final String groupPwd = properties.getProperty(GROUP_PASSWORD_PROPERTY_NAME);
            if (groupPwd != null) {
                xmlConfiguration.setGroupPassword(groupPwd);
            } else {
                throw new MissingPropertyException(GROUP_PASSWORD_PROPERTY_NAME);
            }

            xmlConfiguration.setOverlayMembers(new OverlayMembers());
            boolean isAddressFound = false;
            for (final String propertyName : properties.stringPropertyNames()) {
                if (propertyName.startsWith(PROPERTY_NAME_PREFIX + RegistryOverlayAPI.RegistryItem.ADDRESS)) {
                    isAddressFound = true;
                    final String address = properties.getProperty(propertyName);
                    if (address != null && !address.isEmpty()) {
                        final String[] addressTokens = address.split(":");
                        if (addressTokens.length > 2) {
                            throw new InvalidPropertyException(propertyName, address);
                        } else {
                            final String host = addressTokens[0];
                            if (host.isEmpty()) {
                                throw new InvalidPropertyException(propertyName, address);
                            } else {
                                final OverlayMember member = new OverlayMember();
                                xmlConfiguration.getOverlayMembers().getOverlayMember().add(member);
                                member.setValue(host);

                                if (addressTokens.length == 2) {
                                    final String portStr = addressTokens[1];
                                    if (portStr != null && !portStr.isEmpty()) {
                                        member.setPort(new BigInteger(portStr));
                                    }
                                }
                            }
                        }
                    } else {
                        throw new InvalidPropertyException(propertyName, address);
                    }
                }
            }
            if (!isAddressFound) {
                throw new MissingPropertyException(MEMBER_HOST_PROPERTY_NAME_PATTERN);
            }

            try {
                final Document document = DocumentBuilders.newDocument();
                ConfigurationBuilder.marshallConfiguration(xmlConfiguration, document);
                return document;
            } catch (final ConfigurationException e) {
                throw new MutableImplException(e);
            }
        } else {
            return null;
        }
    }

    @Override
    public Document retrieveEndpointDescription(final QName serviceName, final String endpointName)
            throws SharedAreaException {
        return toDocument(getDescription(serviceName, endpointName));
    }

    private static Document toDocument(final String desc) throws SharedAreaException {
        try {
            return desc != null ? XMLHelper.createDocumentFromString(desc) : null;
        } catch (final IOException | SAXException e) {
            throw new SharedAreaException("Can't parse the descriptions stored in the registry", e);
        }
    }
    
    @Override
    public Optional<Document> getDescription(final ServiceEndpoint endpoint) throws SharedAreaException {
        try {
            if (!(endpoint instanceof InternalEndpointImpl) || !this.endpointsMap.containsKey(
                    getEndpointKey(endpoint.getServiceName(), endpoint.getEndpointName(), Type.INTERNAL))) {
                // for these cases, the endpoint is invalid (either not internal or do not exists anymore)
                return null;
            } else {
                return Optional.fromNullable(((InternalEndpointImpl) endpoint).getDescription(this));
            }
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public String getDescription(final QName serviceName, final String endpointName) throws SharedAreaException {
        try {
            return this.endpointDescriptionsMap.get(getEndpointKey(serviceName, endpointName, Type.INTERNAL));
        } catch (final RuntimeException e) {
            throw new SharedAreaException(e);
        }
    }

    @Override
    public String getImplementationName() {
        return "Petals Distributed Shared Memory Overlay Client";
    }

    @Override
    public String getImplementationVersion() {
        return this.getClass().getPackage().getImplementationVersion();
    }

    private enum Stored2Petals implements Function<StoredRegistryEndpoint, PetalsServiceEndpoint> {
        INSTANCE;

        @Override
        public PetalsServiceEndpoint apply(final StoredRegistryEndpoint input) {
            assert input != null;
            switch (input.getType()) {
                case INTERNAL:
                    return new InternalEndpointImpl(input);
                case EXTERNAL:
                    return new ExternalEndpointImpl(input);
                default:
                    throw new IllegalArgumentException("impossible case");
            }
        }
    }

    private static Type toStoredType(final EndpointType type) {
        switch (type) {
            case EXTERNAL:
                return Type.EXTERNAL;
            case INTERNAL:
                return Type.INTERNAL;
            default:
                throw new IllegalArgumentException("impossible case");
        }
    }

    private static StoredRegistryEndpoint toStoredEndpoint(final Type type, final ServiceEndpoint endpoint,
            final Location location) {
        return toStoredEndpoint(type, endpoint.getEndpointName(), endpoint.getServiceName(), endpoint.getInterfaces(),
                location);
    }

    private static StoredRegistryEndpoint toStoredEndpoint(final Type type, final String endpointName,
            final QName serviceName, final QName[] interfaces, final Location location) {
        final int l = interfaces.length;
        final String[] interfacesString = new String[l];
        for (int i = 0; i < l; i++) {
            interfacesString[i] = interfaces[i].toString();
        }
        return new StoredRegistryEndpoint(type, endpointName, serviceName.toString(), interfacesString,
                location.getContainerName(), location.getComponentName());
    }

    private static Predicate<?, ?> matchService(final QName serviceName) {
        return new SqlPredicate(String.format(DataModelApi.ENDPOINT_SERVICE_INDEX + " = '%s'", serviceName.toString()));
    }

    private static Predicate<?, ?> matchInterface(final QName interfaceName) {
        return new SqlPredicate(
                String.format(DataModelApi.ENDPOINT_INTERFACES_INDEX + " = '%s'", interfaceName.toString()));
    }

    private static Predicate<?, ?> matchType(final EndpointType type) {
        return Predicates.equal(DataModelApi.ENDPOINT_TYPE_INDEX, toStoredType(type).name());
    }
}
