/**
 * Copyright (c) 2013-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import java.util.Set;

import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.ow2.petals.microkernel.api.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.util.ConfigurationServiceUtils;
import org.ow2.petals.microkernel.configuration.ConfigurationTestUtils;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

import com.ebmwebsourcing.easycommons.log.LoggingUtil;
import com.hazelcast.client.HazelcastClient;
import com.hazelcast.client.config.ClientConfig;
import com.hazelcast.client.config.ClientNetworkConfig;
import com.hazelcast.config.GroupConfig;
import com.hazelcast.core.HazelcastInstance;
import com.hazelcast.core.IMap;

/**
 * Unit tests of {@link DynamicTopologyListener}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class DynamicTopologyListenerTest extends AbstractRegistryOverlayImplTest {

    private final static String CONTAINER_NAME_1 = "sample-1";

    private final static String CONTAINER_NAME_2 = "sample-2";

    private final static String CONTAINER_DESCRIPTION_1 = "description of the container 1";

    private final static String CONTAINER_DESCRIPTION_2 = "description of the container 2";

    private final static int CONTAINER_TCP_PORT_1 = 7801;

    private final static int CONTAINER_TCP_PORT_2 = 7802;

    private final LoggingUtil log = new LoggingUtil(
            java.util.logging.Logger.getLogger(DynamicTopologyListenerTest.class.getSimpleName()));

    private HazelcastInstance hzlClient;

    @Before
    public void before() {
        final ClientConfig hzlConfig = new ClientConfig();
        final ClientNetworkConfig networkConfig = new ClientNetworkConfig();
        networkConfig.addAddress(PETALS_REGISTRY_NODE.getHost() + ":" + PETALS_REGISTRY_NODE.getPort());
        hzlConfig.setNetworkConfig(networkConfig);
        final GroupConfig groupConfig = new GroupConfig();
        hzlConfig.setGroupConfig(groupConfig);
        groupConfig.setName(PETALS_REGISTRY_NODE.getGroupName());
        groupConfig.setPassword(PETALS_REGISTRY_NODE.getPassword());
        this.hzlClient = HazelcastClient.newHazelcastClient(hzlConfig);
    }

    @After
    public void after() {
        this.hzlClient.shutdown();
    }

    /**
     * <p>
     * Check {@link DynamicTopologyListener} against adding and removing container:
     * </p>
     * <ul>
     * <li>A basic sample topology is added to the distributed map,</li>
     * <li>A container is added to the initial topology.</li>
     * </ul>
     * <p>
     * Expected results:
     * </p>
     * <ul>
     * <li>the basic sample topology is correctly added to the distributed map,</li>
     * <li>when the topology is updated adding a new container, it is correctly added to the configuration service,</li>
     * </ul>
     */
    @Test
    public void testEntryUpdated_ContainerAdded() throws InterruptedException {
        final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();
        final ContainerConfiguration currentContainerConfig = new ContainerConfiguration();
        ConfigurationServiceUtils.initializeContainerConfiguration(currentContainerConfig,
                initialTopology.getDomain().getContainer().get(0));

        final Container anotherInitialContainer = new Container();
        initialTopology.getDomain().getContainer().add(anotherInitialContainer);
        anotherInitialContainer.setName(CONTAINER_NAME_1);
        anotherInitialContainer.setDescription(CONTAINER_DESCRIPTION_1);
        anotherInitialContainer.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        anotherInitialContainer.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        anotherInitialContainer.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);
        final JmxService jmxService0 = new JmxService();
        anotherInitialContainer.setJmxService(jmxService0);
        final TransportService transportService0 = new TransportService();
        transportService0.setTcpPort(CONTAINER_TCP_PORT_1);
        anotherInitialContainer.setTransportService(transportService0);
        final ContainerConfiguration anotherInitialContainerConfig = new ContainerConfiguration();
        ConfigurationServiceUtils.initializeContainerConfiguration(anotherInitialContainerConfig,
                anotherInitialContainer);

        final DynamicTopologyListenerTestCfgSvc dtltcs = new DynamicTopologyListenerTestCfgSvc(currentContainerConfig,
                new ContainerConfiguration[] { anotherInitialContainerConfig });
        final DynamicTopologyListener dtl = new DynamicTopologyListener(dtltcs, ConfigurationTestUtils.DOMAIN_NAME,
                this.log);

        final IMap<String, Topology> topologies = this.hzlClient.getMap("testEntryUpdated_ContainerAdded");
        topologies.addEntryListener(dtl, ConfigurationTestUtils.DOMAIN_NAME, true);

        topologies.put(ConfigurationTestUtils.DOMAIN_NAME, initialTopology);

        // Assert test environment
        final Set<ContainerConfiguration> containers_1 = dtltcs.getContainersConfiguration();
        assertNotNull("Container list is null", containers_1);
        assertEquals("Unexpected remote container list size", 1, containers_1.size());
        assertEquals("Unexpected remote container", CONTAINER_NAME_1,
                containers_1.toArray(new ContainerConfiguration[containers_1.size()])[0].getName());

        // Execute test: Add a container to the topology
        final Topology topology_2 = ConfigurationTestUtils.createBasicSampleTopology();
        final Container container2_1 = new Container();
        topology_2.getDomain().getContainer().add(anotherInitialContainer);
        topology_2.getDomain().getContainer().add(container2_1);
        container2_1.setName(CONTAINER_NAME_2);
        container2_1.setDescription(CONTAINER_DESCRIPTION_2);
        container2_1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        container2_1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        container2_1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);
        final JmxService jmxService1 = new JmxService();
        container2_1.setJmxService(jmxService1);
        final TransportService transportService1 = new TransportService();
        transportService1.setTcpPort(CONTAINER_TCP_PORT_2);
        container2_1.setTransportService(transportService1);

        topologies.put(ConfigurationTestUtils.DOMAIN_NAME, topology_2);
        synchronized (dtltcs) {
            dtltcs.wait(10000);
        }

        // Assert test results
        final Set<ContainerConfiguration> containers_2 = dtltcs.getContainersConfiguration();
        assertNotNull("Container list is null", containers_2);
        assertEquals("Unexpected container list size", 2, containers_2.size());
        boolean container1Found = false;
        boolean container2Found = false;
        for (final ContainerConfiguration containerConfiguration : containers_2) {
            if (containerConfiguration.getName().equals(CONTAINER_NAME_1)) {
                container1Found = true;
            } else if (containerConfiguration.getName().equals(CONTAINER_NAME_2)) {
                container2Found = true;
            } else {
                fail("Unexpected container: " + containerConfiguration.getName());
            }
        }
        assertTrue("Container " + CONTAINER_NAME_1 + " not found", container1Found);
        assertTrue("Container " + CONTAINER_NAME_2 + " not found", container2Found);
    }

    /**
     * <p>
     * Check {@link DynamicTopologyListener} against adding and removing container:
     * </p>
     * <ul>
     * <li>A basic sample topology with 2 containers is added to the distributed map,</li>
     * <li>The remote container is removed from the initial topology.</li>
     * </ul>
     * <p>
     * Expected results:
     * </p>
     * <ul>
     * <li>the initial topology is correctly added to the distributed map,</li>
     * <li>when the topology is updated removing the remote container, it is correctly removed from the configuration
     * service,</li>
     * </ul>
     */
    @Test
    public void testEntryUpdated_ContainerRemoved() throws InterruptedException {
        final Topology initialTopology = ConfigurationTestUtils.createBasicSampleTopology();
        final ContainerConfiguration currentContainerConfig = new ContainerConfiguration();
        ConfigurationServiceUtils.initializeContainerConfiguration(currentContainerConfig,
                initialTopology.getDomain().getContainer().get(0));

        final Container remoteContainer1 = new Container();
        remoteContainer1.setName(CONTAINER_NAME_1);
        remoteContainer1.setDescription(CONTAINER_DESCRIPTION_1);
        remoteContainer1.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        remoteContainer1.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        remoteContainer1.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);
        final JmxService jmxService1 = new JmxService();
        remoteContainer1.setJmxService(jmxService1);
        final TransportService transportService1 = new TransportService();
        transportService1.setTcpPort(CONTAINER_TCP_PORT_1);
        remoteContainer1.setTransportService(transportService1);
        final ContainerConfiguration remoteContainerConfig1 = new ContainerConfiguration();
        ConfigurationServiceUtils.initializeContainerConfiguration(remoteContainerConfig1, remoteContainer1);

        final Container remoteContainer2 = new Container();
        remoteContainer2.setName(CONTAINER_NAME_2);
        remoteContainer2.setDescription(CONTAINER_DESCRIPTION_2);
        remoteContainer2.setHost(ConfigurationTestUtils.CONTAINER_HOST);
        remoteContainer2.setUser(ConfigurationTestUtils.CONTAINER_JMX_USER);
        remoteContainer2.setPassword(ConfigurationTestUtils.CONTAINER_JMX_PWD);
        final JmxService jmxService2 = new JmxService();
        remoteContainer2.setJmxService(jmxService2);
        final TransportService transportService2 = new TransportService();
        transportService2.setTcpPort(CONTAINER_TCP_PORT_2);
        remoteContainer2.setTransportService(transportService2);
        final ContainerConfiguration remoteContainerConfig2 = new ContainerConfiguration();
        ConfigurationServiceUtils.initializeContainerConfiguration(remoteContainerConfig2, remoteContainer2);

        final DynamicTopologyListenerTestCfgSvc dtltcs = new DynamicTopologyListenerTestCfgSvc(currentContainerConfig,
                new ContainerConfiguration[] { remoteContainerConfig1, remoteContainerConfig2 });
        final DynamicTopologyListener dtl = new DynamicTopologyListener(dtltcs, ConfigurationTestUtils.DOMAIN_NAME,
                this.log);

        final IMap<String, Topology> topologies = hzlClient.getMap("testEntryUpdated_ContainerRemoved");
        topologies.addEntryListener(dtl, ConfigurationTestUtils.DOMAIN_NAME, true);

        initialTopology.getDomain().getContainer().add(remoteContainer1);
        initialTopology.getDomain().getContainer().add(remoteContainer2);
        topologies.put(ConfigurationTestUtils.DOMAIN_NAME, initialTopology);

        // Assert test environment
        final Set<ContainerConfiguration> containers_1 = dtltcs.getContainersConfiguration();
        assertNotNull("Container list is null", containers_1);
        assertEquals("Unexpected container list size", 2, containers_1.size());
        boolean container1Found = false;
        boolean container2Found = false;
        for (final ContainerConfiguration containerConfiguration : containers_1) {
            if (containerConfiguration.getName().equals(CONTAINER_NAME_1)) {
                container1Found = true;
            } else if (containerConfiguration.getName().equals(CONTAINER_NAME_2)) {
                container2Found = true;
            } else {
                fail("Unexpected container: " + containerConfiguration.getName());
            }
        }
        assertTrue("Container " + CONTAINER_NAME_1 + " not found", container1Found);
        assertTrue("Container " + CONTAINER_NAME_2 + " not found", container2Found);

        // Execute test: Remove the remote container from the topology
        final Topology topology_2 = ConfigurationTestUtils.createBasicSampleTopology();
        topology_2.getDomain().getContainer().add(remoteContainer2);
        topologies.put(ConfigurationTestUtils.DOMAIN_NAME, topology_2);
        synchronized (dtltcs) {
            dtltcs.wait(10000);
        }

        // Assert test results
        final Set<ContainerConfiguration> containers_2 = dtltcs.getContainersConfiguration();
        assertNotNull("Container list is null", containers_2);
        assertEquals("Unexpected container list size", 1, containers_2.size());
        assertEquals("Unexpected remote container", CONTAINER_NAME_2,
                containers_2.toArray(new ContainerConfiguration[containers_2.size()])[0].getName());
    }
}
