/**
 * Copyright (c) 2013-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import java.util.Collection;

import javax.xml.namespace.QName;

import org.junit.Test;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.microkernel.api.communication.sharedarea.SharedAreaService;
import org.ow2.petals.registry_overlay.api.Constants;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.ebmwebsourcing.easycommons.uuid.QualifiedUUIDGenerator;
import com.hazelcast.config.Config;
import com.hazelcast.config.NetworkConfig;
import com.hazelcast.core.Hazelcast;

/**
 * Unit tests of
 * {@link RegistryOverlayImpl#registerInternalEndpoint(String, QName, QName[], String, org.w3c.dom.Document)}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class RegistryOverlayImpl_ActivateEndpoint_Test extends AbstractRegistryOverlayImplTest {

    /**
     * A namespace name
     */
    private final static String A_NAMESPACE = "http://petals.ow2.org/unit-tests";

    /**
     * A service name
     */
    private final static QName SERVICE_NAME_1 = new QName(A_NAMESPACE, "service-1");

    /**
     * A interface name
     */
    private final static QName INTERFACE_NAME_1 = new QName(A_NAMESPACE, "interface-1");

    /**
     * A endpoint name
     */
    private final static String ENDPOINT_NAME_1;

    /**
     * A container name
     */
    private final static String CONTAINER_NAME_1 = "container-1";

    static {
        final QualifiedUUIDGenerator uuidGenerator = new QualifiedUUIDGenerator("petals");
        ENDPOINT_NAME_1 = uuidGenerator.getNewID();
    }

    /**
     * <p>
     * Check the method
     * {@link RegistryOverlayImpl#registerInternalEndpoint(String, QName, QName[], String, org.w3c.dom.Document)}:
     * </p>
     * <ol>
     * <li>in a dedicated topology,</li>
     * <li>activate only one valid endpoint.</li>
     * </ol>
     * <p>
     * Expected result:
     * </p>
     * <ol>
     * <li>no error occurs, the endpoint is correctly registered</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpoint(QName, String, EndpointType)}</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpointsForService(QName, EndpointType)}</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpointsForInterface(QName, EndpointType)}</li>
     * </ol>
     */
    @Test
    public void registerOneValidEndpoint() throws Exception {

        final RegistryOverlayImpl registryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(registryOverlayImpl, "configurationService",
                RegistryOverlayImpl_ActivateEndpoint_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), this.getClass().getSimpleName()
                                + "_registerOneValidEndpoint", CONTAINER_NAME_1));
        try {
            registryOverlayImpl.start();

            // Add the endpoint
            registryOverlayImpl.registerInternalEndpoint(ENDPOINT_NAME_1, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", null);

            // Retrieve the endpoint
            final PetalsServiceEndpoint retrievedEndpoint = registryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                    ENDPOINT_NAME_1, EndpointType.INTERNAL);
            assertNotNull(retrievedEndpoint);
            assertEquals(retrievedEndpoint.getServiceName(), SERVICE_NAME_1);
            assertEquals(retrievedEndpoint.getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the service
            final Collection<PetalsServiceEndpoint> svcEpForSvc = registryOverlayImpl
                    .getEndpointsForService(SERVICE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForSvc);
            assertEquals("No endpoint retrieved for service", svcEpForSvc.size(), 1);
            assertEquals(svcEpForSvc.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the interface
            final Collection<PetalsServiceEndpoint> svcEpForItf = registryOverlayImpl
                    .getEndpointsForInterface(INTERFACE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForItf);
            assertEquals("No endpoint retrieved for interface", svcEpForItf.size(), 1);
            assertEquals(svcEpForItf.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

        } finally {
            registryOverlayImpl.stop();
        }
    }

    /**
     * <p>
     * Check the method
     * {@link RegistryOverlayImpl#registerInternalEndpoint(String, QName, QName[], String, org.w3c.dom.Document)}:
     * </p>
     * <ol>
     * <li>register the same endpoint in two different topologies,</li>
     * <li>both topologies use the same shared area.</li>
     * </ol>
     * <p>
     * Expected result:
     * </p>
     * <ol>
     * <li>no error occurs, the endpoint is correctly registered in each topology</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpoint(QName, String, EndpointType)}</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpointsForService(QName, EndpointType)}</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpointsForInterface(QName, EndpointType)}</li>
     * </ol>
     */
    @Test
    public void registerTwiceTheSameValidEndpointIn2Topologies() throws Exception {

        // Create a second cluster member in memory on a dedicated port. It will be shutdown by
        // AbstractRegistryOverlayImplTest.stopHazelcastCluster
        final Config hzlConfig = new Config();
        final NetworkConfig networkConfig = new NetworkConfig();
        hzlConfig.setNetworkConfig(networkConfig);
        networkConfig.setPort(Constants.DEFAULT_PORT);
        Hazelcast.newHazelcastInstance(hzlConfig);

        final RegistryOverlayImpl registryOverlayImpl_1 = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(registryOverlayImpl_1, "configurationService",
                RegistryOverlayImpl_ActivateEndpoint_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), this.getClass().getSimpleName()
                                + "_registerTwiceTheSameValidEndpointIn2Topologies_1", CONTAINER_NAME_1));

        final RegistryOverlayImpl registryOverlayImpl_2 = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(registryOverlayImpl_2, "configurationService",
                RegistryOverlayImpl_ActivateEndpoint_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), this.getClass().getSimpleName()
                                + "_registerTwiceTheSameValidEndpointIn2Topologies_2", CONTAINER_NAME_1));

        try {
            registryOverlayImpl_1.start();
            registryOverlayImpl_2.start();

            // Add the endpoint int topology 1
            registryOverlayImpl_1.registerInternalEndpoint(ENDPOINT_NAME_1, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", null);
            // Add the endpoint int topology 2
            registryOverlayImpl_2.registerInternalEndpoint(ENDPOINT_NAME_1, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", null);

            // Retrieve the endpoint in topology 1
            final PetalsServiceEndpoint retrievedEndpoint_1 = registryOverlayImpl_1.getEndpoint(SERVICE_NAME_1,
                    ENDPOINT_NAME_1, EndpointType.INTERNAL);
            assertNotNull(retrievedEndpoint_1);
            assertEquals(retrievedEndpoint_1.getServiceName(), SERVICE_NAME_1);
            assertEquals(retrievedEndpoint_1.getEndpointName(), ENDPOINT_NAME_1);
            // Retrieve the endpoint in topology 2
            final PetalsServiceEndpoint retrievedEndpoint_2 = registryOverlayImpl_2.getEndpoint(SERVICE_NAME_1,
                    ENDPOINT_NAME_1, EndpointType.INTERNAL);
            assertNotNull(retrievedEndpoint_2);
            assertEquals(retrievedEndpoint_2.getServiceName(), SERVICE_NAME_1);
            assertEquals(retrievedEndpoint_2.getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the
            // service in topology 1
            final Collection<PetalsServiceEndpoint> svcEpForSvc_1 = registryOverlayImpl_1
                    .getEndpointsForService(SERVICE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForSvc_1);
            assertEquals("No endpoint retrieved for service", svcEpForSvc_1.size(), 1);
            assertEquals(svcEpForSvc_1.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the
            // service in topology 2
            final Collection<PetalsServiceEndpoint> svcEpForSvc_2 = registryOverlayImpl_2
                    .getEndpointsForService(SERVICE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForSvc_2);
            assertEquals("No endpoint retrieved for service", svcEpForSvc_2.size(), 1);
            assertEquals(svcEpForSvc_2.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the
            // interface in topology 1
            final Collection<PetalsServiceEndpoint> svcEpForItf_1 = registryOverlayImpl_1
                    .getEndpointsForInterface(INTERFACE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForItf_1);
            assertEquals("No endpoint retrieved for interface", svcEpForItf_1.size(), 1);
            assertEquals(svcEpForItf_1.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the
            // interface in topology 2
            final Collection<PetalsServiceEndpoint> svcEpForItf_2 = registryOverlayImpl_2
                    .getEndpointsForInterface(INTERFACE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForItf_2);
            assertEquals("No endpoint retrieved for interface", svcEpForItf_2.size(), 1);
            assertEquals(svcEpForItf_2.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

        } finally {
            registryOverlayImpl_1.stop();
            registryOverlayImpl_2.stop();
        }
    }

    /**
     * <p>
     * Check the method
     * {@link RegistryOverlayImpl#registerInternalEndpoint(String, QName, QName[], String, org.w3c.dom.Document)}:
     * </p>
     * <ol>
     * <li>register the same endpoint twice in the same topology</li>
     * </ol>
     * <p>
     * Expected result:
     * </p>
     * <ol>
     * <li>the endpoint is correctly registered on the first one,</li>
     * <li>the error EndpointAlreadyExistsException is thrown on the second registration,</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpoint(QName, String, EndpointType)}</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpointsForService(QName, EndpointType)}</li>
     * <li>the endpoint is retrieved using {@link SharedAreaService#getEndpointsForInterface(QName, EndpointType)}</li>
     * </ol>
     */
    @Test
    public void registerTwiceTheSameValidEndpointInTheSameTopology() throws Exception {

        final RegistryOverlayImpl registryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(registryOverlayImpl, "configurationService",
                RegistryOverlayImpl_ActivateEndpoint_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), this.getClass().getSimpleName()
                                + "_registerTwiceTheSameValidEndpointInTheSameTopology", CONTAINER_NAME_1));
        try {
            registryOverlayImpl.start();

            // 1st Add of the endpoint
            registryOverlayImpl.registerInternalEndpoint(ENDPOINT_NAME_1, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", null);

            // Retrieve the endpoint
            final PetalsServiceEndpoint retrievedEndpoint = registryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                    ENDPOINT_NAME_1, EndpointType.INTERNAL);
            assertNotNull(retrievedEndpoint);
            assertEquals(retrievedEndpoint.getServiceName(), SERVICE_NAME_1);
            assertEquals(retrievedEndpoint.getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the service
            final Collection<PetalsServiceEndpoint> svcEpForSvc = registryOverlayImpl
                    .getEndpointsForService(SERVICE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForSvc);
            assertEquals("No endpoint retrieved for service", svcEpForSvc.size(), 1);
            assertEquals(svcEpForSvc.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the interface
            final Collection<PetalsServiceEndpoint> svcEpForItf = registryOverlayImpl
                    .getEndpointsForInterface(INTERFACE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForItf);
            assertEquals("No endpoint retrieved for interface", svcEpForItf.size(), 1);
            assertEquals(svcEpForItf.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

            // 2nd Add of the endpoint (it should fail)
            assertNull(registryOverlayImpl.registerInternalEndpoint(ENDPOINT_NAME_1, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", null));

            // Retrieve the endpoint
            final PetalsServiceEndpoint retrievedEndpoint2 = registryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                    ENDPOINT_NAME_1, EndpointType.INTERNAL);
            assertNotNull(retrievedEndpoint2);
            assertEquals(retrievedEndpoint2.getServiceName(), SERVICE_NAME_1);
            assertEquals(retrievedEndpoint2.getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the service
            final Collection<PetalsServiceEndpoint> svcEpForSvc2 = registryOverlayImpl
                    .getEndpointsForService(SERVICE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForSvc2);
            assertEquals("No endpoint retrieved for service", svcEpForSvc2.size(), 1);
            assertEquals(svcEpForSvc2.iterator().next().getEndpointName(), ENDPOINT_NAME_1);

            // Check that only one endpoint is registered as implementing the interface
            final Collection<PetalsServiceEndpoint> svcEpForItf2 = registryOverlayImpl
                    .getEndpointsForInterface(INTERFACE_NAME_1, null);
            assertNotNull("Null list returned", svcEpForItf2);
            assertEquals("No endpoint retrieved for interface", svcEpForItf2.size(), 1);
            assertEquals(svcEpForItf2.iterator().next().getEndpointName(), ENDPOINT_NAME_1);


        } finally {
            registryOverlayImpl.stop();
        }
    }

}
