/**
 * Copyright (c) 2013-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.junit.Test;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.ow2.petals.jbi.servicedesc.endpoint.impl.JBIServiceEndpointImpl;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.ebmwebsourcing.easycommons.uuid.QualifiedUUIDGenerator;

/**
 * Unit tests of {@link RegistryOverlayImpl#getEndpoint(QName, String, EndpointType)}.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class RegistryOverlayImpl_GetEndpoint_Test extends AbstractRegistryOverlayImplTest {

    /**
     * A namespace name
     */
    private final static String A_NAMESPACE = "http://petals.ow2.org/unit-tests";

    /**
     * A service name
     */
    private final static QName SERVICE_NAME_1 = new QName(A_NAMESPACE, "service-1");

    /**
     * A interface name
     */
    private final static QName INTERFACE_NAME_1 = new QName(A_NAMESPACE, "interface-1");

    /**
     * A local container name
     */
    private final static String LOCAL_CONTAINER_NAME = "local-container";

    /**
     * An internal endpoint name on a local container
     */
    private final static String LOCAL_ENDPOINT_NAME_1;

    static {
        final QualifiedUUIDGenerator uuidGenerator = new QualifiedUUIDGenerator("petals");
        LOCAL_ENDPOINT_NAME_1 = uuidGenerator.getNewID();
    }

    /**
     * <p>
     * Check the method {@link RegistryOverlayImpl#getEndpoint(QName, String, EndpointType)} :
     * </p>
     * <ol>
     * <li>in a dedicated topology,</li>
     * <li>an internal endpoint is registered on a local container.</li>
     * </ol>
     * <p>
     * Expected result:
     * </p>
     * <ol>
     * <li>no error occurs</li>
     * <li>the endpoint is retrieved with the right attributes</li>
     * </ol>
     */
    @Test
    public void getEndpointInternal() throws Exception {

        final String topologyName = this.getClass().getSimpleName() + "_getEndpoint";
        final RegistryOverlayImpl localRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(localRegistryOverlayImpl, "configurationService",
                RegistryOverlayImpl_GetEndpoint_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), topologyName, LOCAL_CONTAINER_NAME));

        try {
            localRegistryOverlayImpl.start();

            // Read the wsdl
            final Document document = readEndpointDescription("echo.wsdl");

            // Add endpoint
            localRegistryOverlayImpl.registerInternalEndpoint(LOCAL_ENDPOINT_NAME_1, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", document);

            // Check that endpoint is registered
            final PetalsServiceEndpoint registeredEp = localRegistryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                    LOCAL_ENDPOINT_NAME_1, EndpointType.INTERNAL);
            assertNotNull("Null list returned", registeredEp);
            assertEquals("Unexpected service name", SERVICE_NAME_1, registeredEp.getServiceName());
            assertEquals("Unexpected endpoint name", LOCAL_ENDPOINT_NAME_1, registeredEp.getEndpointName());
            assertEndpointDescriptionParsable(localRegistryOverlayImpl, registeredEp);

            // ensure we are not getting things mixed up between internal and external
            final PetalsServiceEndpoint retrievedEndpoint2 = localRegistryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                    LOCAL_ENDPOINT_NAME_1, EndpointType.EXTERNAL);
            assertNull(retrievedEndpoint2);

        } finally {
            localRegistryOverlayImpl.stop();
        }
    }

    /**
     * <p>
     * Check the method {@link RegistryOverlayImpl#getEndpoint(QName, String, EndpointType)} :
     * </p>
     * <ol>
     * <li>in a dedicated topology,</li>
     * <li>an external endpoint is registered on a local container.</li>
     * </ol>
     * <p>
     * Expected result:
     * </p>
     * <ol>
     * <li>no error occurs</li>
     * <li>the endpoint is retrieved with the right attributes</li>
     * </ol>
     */
    @Test
    public void getEndpointExternal() throws Exception {

        final String topologyName = this.getClass().getSimpleName() + "_getEndpoint";
        final RegistryOverlayImpl localRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(localRegistryOverlayImpl, "configurationService",
                RegistryOverlayImpl_GetEndpoint_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), topologyName, LOCAL_CONTAINER_NAME));

        try {
            localRegistryOverlayImpl.start();

            final ServiceEndpoint ext = new JBIServiceEndpointImpl(EndpointType.EXTERNAL, LOCAL_ENDPOINT_NAME_1,
                    SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, new Location("", ""));

            // Add endpoint
            localRegistryOverlayImpl.registerExternalEndpoint(ext, "");

            // Check that endpoint is registered
            final PetalsServiceEndpoint registeredEp = localRegistryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                    LOCAL_ENDPOINT_NAME_1, EndpointType.EXTERNAL);
            assertNotNull("Null list returned", registeredEp);
            assertEquals("Unexpected service name", SERVICE_NAME_1, registeredEp.getServiceName());
            assertEquals("Unexpected endpoint name", LOCAL_ENDPOINT_NAME_1, registeredEp.getEndpointName());

            // ensure we are not getting things mixed up between internal and external
            final PetalsServiceEndpoint retrievedEndpoint2 = localRegistryOverlayImpl.getEndpoint(SERVICE_NAME_1,
                    LOCAL_ENDPOINT_NAME_1, EndpointType.INTERNAL);
            assertNull(retrievedEndpoint2);
        } finally {
            localRegistryOverlayImpl.stop();
        }
    }
}
