/**
 * Copyright (c) 2013-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.Collection;

import javax.xml.namespace.QName;

import org.junit.Test;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;
import org.w3c.dom.Document;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.ebmwebsourcing.easycommons.uuid.QualifiedUUIDGenerator;

/**
 * Unit tests of {@link RegistryOverlayImpl#getEndpointsForInterface(QName, EndpointType)}.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class RegistryOverlayImpl_GetEndpointsForInterface_Test extends AbstractRegistryOverlayImplTest {

    /**
     * A namespace name
     */
    private final static String A_NAMESPACE = "http://petals.ow2.org/unit-tests";

    /**
     * A service name
     */
    private final static QName SERVICE_NAME_1 = new QName(A_NAMESPACE, "service-1");

    /**
     * Another service name
     */
    private final static QName SERVICE_NAME_2 = new QName(A_NAMESPACE, "service-2");

    /**
     * A interface name
     */
    private final static QName INTERFACE_NAME_1 = new QName(A_NAMESPACE, "interface-1");

    /**
     * Another interface name
     */
    private final static QName INTERFACE_NAME_2 = new QName(A_NAMESPACE, "interface-2");

    /**
     * Another interface name
     */
    private final static QName INTERFACE_NAME_3 = new QName(A_NAMESPACE, "interface-3");

    /**
     * A local container name
     */
    private final static String LOCAL_CONTAINER_NAME = "local-container";

    /**
     * An internal endpoint name on a local container
     */
    private final static String LOCAL_INTERNAL_ENDPOINT_NAME_1;

    /**
     * Another internal endpoint name on a local container
     */
    private final static String LOCAL_INTERNAL_ENDPOINT_NAME_2;

    /**
     * Another internal endpoint name on a local container
     */
    private final static String LOCAL_INTERNAL_ENDPOINT_NAME_3;

    /**
     * An external endpoint name on a local container
     */
    private final static String LOCAL_EXTERNAL_ENDPOINT_NAME_1;

    static {
        final QualifiedUUIDGenerator uuidGenerator = new QualifiedUUIDGenerator("petals");
        LOCAL_INTERNAL_ENDPOINT_NAME_1 = uuidGenerator.getNewID();
        LOCAL_INTERNAL_ENDPOINT_NAME_2 = uuidGenerator.getNewID();
        LOCAL_INTERNAL_ENDPOINT_NAME_3 = uuidGenerator.getNewID();
        LOCAL_EXTERNAL_ENDPOINT_NAME_1 = uuidGenerator.getNewID();
    }

    /**
     * <p>
     * Check the method {@link RegistryOverlayImpl#getEndpointsForInterface(QName, EndpointType)} :
     * </p>
     * <ol>
     * <li>in a dedicated topology,</li>
     * <li>1 internal endpoint EP1 is registered as implementing interface A,</li>
     * <li>2 internal endpoints EP2 and EP3 are registered as implementing interface B,</li>
     * <li>1 external endpoint EP4 is registered as implementing interface B</li>
     * </ol>
     * <p>
     * Expected result:
     * </p>
     * <ol>
     * <li>no error occurs</li>
     * <li>requesting with interface A and without endpoint type: only EP1 is returned,</li>
     * <li>requesting with interface B and without endpoint type: only EP2, EP3 and EP4 are returned,</li>
     * <li>requesting with interface B and endpoint type external: only EP4 is returned,</li>
     * <li>requesting with interface B and endpoint type internal: only EP2 and EP3 are returned,</li>
     * <li>requesting all internal endpoints: only EP1, EP2 and EP3 are returned,</li>
     * <li>requesting all external endpoints: only EP4 is returned</li>
     * </ol>
     */
    @Test
    public void getEndpointsForInterface() throws Exception {

        final String topologyName = this.getClass().getSimpleName() + "_getEndpointsForInterface";
        final RegistryOverlayImpl localRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(localRegistryOverlayImpl, "configurationService",
                RegistryOverlayImpl_GetEndpointsForInterface_Test.createCfgServiceMock(
                        PETALS_REGISTRY_NODE.getGroupName(), PETALS_REGISTRY_NODE.getPassword(),
                        PETALS_REGISTRY_NODE.getHost(), String.valueOf(PETALS_REGISTRY_NODE.getPort()), topologyName,
                        LOCAL_CONTAINER_NAME));

        try {
            localRegistryOverlayImpl.start();

            // Read the wsdl
            final Document document = readEndpointDescription("echo.wsdl");

            // Add endpoints
            localRegistryOverlayImpl.registerInternalEndpoint(LOCAL_INTERNAL_ENDPOINT_NAME_1, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", document);

            localRegistryOverlayImpl.registerInternalEndpoint(LOCAL_INTERNAL_ENDPOINT_NAME_2, SERVICE_NAME_2,
                    new QName[] { INTERFACE_NAME_2 }, "", document);

            localRegistryOverlayImpl.registerInternalEndpoint(LOCAL_INTERNAL_ENDPOINT_NAME_3, SERVICE_NAME_2,
                    new QName[] { INTERFACE_NAME_2, INTERFACE_NAME_3 }, "", document);

            localRegistryOverlayImpl.registerExternalEndpoint(
                    new MockEndpoint(LOCAL_EXTERNAL_ENDPOINT_NAME_1, SERVICE_NAME_2, new QName[] { INTERFACE_NAME_2 }),
                    "");

            // Check that endpoints are registered
            final Collection<PetalsServiceEndpoint> registeredEndpoints = localRegistryOverlayImpl.getEndpoints(null);
            assertNotNull("Null list returned", registeredEndpoints);
            assertEquals("Invalid response number retrieving endpoints for service", 4, registeredEndpoints.size());
            boolean endpoint1Found = false;
            boolean endpoint2Found = false;
            boolean endpoint3Found = false;
            boolean endpoint4Found = false;
            for (final PetalsServiceEndpoint registeredEndpoint : registeredEndpoints) {
                if (LOCAL_INTERNAL_ENDPOINT_NAME_1.equals(registeredEndpoint.getEndpointName())) {
                    endpoint1Found = true;
                    assertEquals(1, registeredEndpoint.getInterfaces().length);
                    assertEquals("Unexpected interface name", INTERFACE_NAME_1, registeredEndpoint.getInterfaces()[0]);
                    assertEquals("Unexpected endpoint type", EndpointType.INTERNAL, registeredEndpoint.getType());
                } else if (LOCAL_INTERNAL_ENDPOINT_NAME_2.equals(registeredEndpoint.getEndpointName())) {
                    endpoint2Found = true;
                    assertEquals(1, registeredEndpoint.getInterfaces().length);
                    assertEquals("Unexpected interface name", INTERFACE_NAME_2, registeredEndpoint.getInterfaces()[0]);
                    assertEquals("Unexpected endpoint type", EndpointType.INTERNAL, registeredEndpoint.getType());
                } else if (LOCAL_INTERNAL_ENDPOINT_NAME_3.equals(registeredEndpoint.getEndpointName())) {
                    endpoint3Found = true;
                    assertEquals(2, registeredEndpoint.getInterfaces().length);
                    assertEquals("Unexpected interface name", INTERFACE_NAME_2, registeredEndpoint.getInterfaces()[0]);
                    assertEquals("Unexpected endpoint type", EndpointType.INTERNAL, registeredEndpoint.getType());
                } else if (LOCAL_EXTERNAL_ENDPOINT_NAME_1.equals(registeredEndpoint.getEndpointName())) {
                    endpoint4Found = true;
                    assertEquals(1, registeredEndpoint.getInterfaces().length);
                    assertEquals("Unexpected interface name", INTERFACE_NAME_2, registeredEndpoint.getInterfaces()[0]);
                    assertEquals("Unexpected endpoint type", EndpointType.EXTERNAL, registeredEndpoint.getType());
                } else {
                    fail("Unexpected endpoint name: " + registeredEndpoint.getEndpointName());
                }

                assertEndpointDescriptionParsable(localRegistryOverlayImpl, registeredEndpoint);
            }
            assertTrue("Endpoint 1 missing", endpoint1Found);
            assertTrue("Endpoint 2 missing", endpoint2Found);
            assertTrue("Endpoint 3 missing", endpoint3Found);
            assertTrue("Endpoint 4 missing", endpoint4Found);

            // get all endpoints implementing interface A
            final Collection<PetalsServiceEndpoint> allEndpointsImplementingItfA = localRegistryOverlayImpl
                    .getEndpointsForInterface(INTERFACE_NAME_1, null);
            assertNotNull("Null list returned", allEndpointsImplementingItfA);
            assertEquals("Unexpected number of returned endpoints", 1, allEndpointsImplementingItfA.size());
            final PetalsServiceEndpoint endpointImplementingItfA = allEndpointsImplementingItfA.iterator().next();
            assertEquals("Unexpected endpoint name", LOCAL_INTERNAL_ENDPOINT_NAME_1,
                    endpointImplementingItfA.getEndpointName());
            assertEquals("Unexpecetd service name", SERVICE_NAME_1, endpointImplementingItfA.getServiceName());
            assertEndpointDescriptionParsable(localRegistryOverlayImpl, endpointImplementingItfA);

            // get all endpoints implementing interface B
            final Collection<PetalsServiceEndpoint> allEndpointsImplementingItfB = localRegistryOverlayImpl
                    .getEndpointsForInterface(INTERFACE_NAME_2, null);
            assertNotNull("Null list returned", allEndpointsImplementingItfB);
            assertEquals("Unexpected number of returned endpoints", 3, allEndpointsImplementingItfB.size());
            for (final PetalsServiceEndpoint endpointImplementingItfB : allEndpointsImplementingItfB) {
                assertEndpointDescriptionParsable(localRegistryOverlayImpl, endpointImplementingItfB);
            }

            // get all external endpoints implementing interface B
            final Collection<PetalsServiceEndpoint> allExternalEndpointsImplementingItfB = localRegistryOverlayImpl
                    .getEndpointsForInterface(INTERFACE_NAME_2, EndpointType.EXTERNAL);
            assertNotNull("Null list returned", allExternalEndpointsImplementingItfB);
            assertEquals("Unexpected number of returned endpoints", 1, allExternalEndpointsImplementingItfB.size());
            final PetalsServiceEndpoint externalEndpointImplementingItfB = allExternalEndpointsImplementingItfB
                    .iterator().next();
            assertEquals("Unexpected endpoint name", LOCAL_EXTERNAL_ENDPOINT_NAME_1,
                    externalEndpointImplementingItfB.getEndpointName());
            assertEquals("Unexpecetd service name", SERVICE_NAME_2, externalEndpointImplementingItfB.getServiceName());
            assertEndpointDescriptionParsable(localRegistryOverlayImpl, externalEndpointImplementingItfB);

            // get all internal endpoints implementing interface B
            final Collection<PetalsServiceEndpoint> allInternalEndpointsImplementingItfB = localRegistryOverlayImpl
                    .getEndpointsForInterface(INTERFACE_NAME_2, EndpointType.INTERNAL);
            assertNotNull("Null list returned", allInternalEndpointsImplementingItfB);
            assertEquals("Unexpected number of returned endpoints", 2, allInternalEndpointsImplementingItfB.size());
            for (final PetalsServiceEndpoint internalEndpointImplementingItfB : allInternalEndpointsImplementingItfB) {
                assertEndpointDescriptionParsable(localRegistryOverlayImpl, internalEndpointImplementingItfB);
            }

            // get all external endpoints
            final Collection<PetalsServiceEndpoint> allExternalEndpoints = localRegistryOverlayImpl
                    .getEndpoints(EndpointType.EXTERNAL);
            assertNotNull("Null list returned", allExternalEndpoints);
            assertEquals("Unexpected number of returned endpoints", 1, allExternalEndpoints.size());
            final PetalsServiceEndpoint externalEndpoint = allExternalEndpoints.iterator().next();
            assertEquals("Unexpected endpoint name", LOCAL_EXTERNAL_ENDPOINT_NAME_1, externalEndpoint.getEndpointName());
            assertEquals("Unexpecetd service name", SERVICE_NAME_2, externalEndpoint.getServiceName());
            assertEndpointDescriptionParsable(localRegistryOverlayImpl, externalEndpoint);

            // get all internal endpoints
            final Collection<PetalsServiceEndpoint> allInternalEndpoints = localRegistryOverlayImpl
                    .getEndpoints(EndpointType.INTERNAL);
            assertNotNull("Null list returned", allInternalEndpoints);
            assertEquals("Unexpected number of returned endpoints", 3, allInternalEndpoints.size());
            for (final PetalsServiceEndpoint internalEndpoint : allInternalEndpoints) {
                assertEndpointDescriptionParsable(localRegistryOverlayImpl, internalEndpoint);
            }

        } finally {
            localRegistryOverlayImpl.stop();
        }
    }

}
