/**
 * Copyright (c) 2013-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.registry.overlay;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.util.Collection;

import javax.xml.namespace.QName;

import org.junit.Test;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint;
import org.ow2.petals.jbi.servicedesc.endpoint.PetalsServiceEndpoint.EndpointType;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;
import com.ebmwebsourcing.easycommons.uuid.QualifiedUUIDGenerator;

/**
 * Unit tests of {@link RegistryOverlayImpl#getEndpoints(EndpointType)}.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class RegistryOverlayImpl_GetEndpoints_Test extends AbstractRegistryOverlayImplTest {

    /**
     * A namespace name
     */
    private final static String A_NAMESPACE = "http://petals.ow2.org/unit-tests";

    /**
     * A service name
     */
    private final static QName SERVICE_NAME_1 = new QName(A_NAMESPACE, "service-1");

    /**
     * A interface name
     */
    private final static QName INTERFACE_NAME_1 = new QName(A_NAMESPACE, "interface-1");

    /**
     * A local container name
     */
    private final static String LOCAL_CONTAINER_NAME = "local-container";

    /**
     * A remote container name
     */
    private final static String REMOTE_CONTAINER_NAME = "remote-container";

    /**
     * An internal endpoint name on a local container
     */
    private final static String LOCAL_INTERNAL_ENDPOINT_NAME;

    /**
     * An external endpoint name on a local container
     */
    private final static String LOCAL_EXTERNAL_ENDPOINT_NAME;

    /**
     * An internal endpoint name on a remote container
     */
    private final static String REMOTE_INTERNAL_ENDPOINT_NAME;

    /**
     * An external endpoint name on a remote container
     */
    private final static String REMOTE_EXTERNAL_ENDPOINT_NAME;

    static {
        final QualifiedUUIDGenerator uuidGenerator = new QualifiedUUIDGenerator("petals");
        LOCAL_INTERNAL_ENDPOINT_NAME = uuidGenerator.getNewID();
        LOCAL_EXTERNAL_ENDPOINT_NAME = uuidGenerator.getNewID();
        REMOTE_INTERNAL_ENDPOINT_NAME = uuidGenerator.getNewID();
        REMOTE_EXTERNAL_ENDPOINT_NAME = uuidGenerator.getNewID();
    }

    /**
     * <p>
     * Check the method 
     * {@link RegistryOverlayImpl#getEndpoints(EndpointType)} :
     * </p>
     * <ol>
     * <li>in a dedicated topology,</li>
     * <li>an internal endpoint is registered on a local container,</li>
     * <li>an external endpoint is registered on a local container,</li>
     * <li>an internal endpoint is registered on a remote container,</li>
     * <li>an external endpoint is registered on a remote container,</li>
     * </ol>
     * <p>
     * Expected result:
     * </p>
     * <ol>
     * <li>no error occurs</li>
     * <li>all endpoints are retrieved through {@link RegistryOverlayImpl#getEndpoints(EndpointType)}</li>
     * <li>all internal endpoints are retrieved through {@link RegistryOverlayImpl#getEndpoints(EndpointType)} using
     * {@link EndpointType#INTERNAL} as criteria,</li>
     * <li>all external endpoints are retrieved through {@link RegistryOverlayImpl#getEndpoints(EndpointType)} using
     * {@link EndpointType#EXTERNAL} as criteria.</li>
     * </ol>
     */
    @Test
    public void getEndpoints() throws Exception {

        final String topologyName = this.getClass().getSimpleName() + "_getEndpoints";
        final RegistryOverlayImpl localRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(localRegistryOverlayImpl, "configurationService",
                RegistryOverlayImpl_GetEndpoints_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), topologyName, LOCAL_CONTAINER_NAME));

        final RegistryOverlayImpl remoteRegistryOverlayImpl = new RegistryOverlayImpl();
        ReflectionHelper.setPrivateField(remoteRegistryOverlayImpl, "configurationService",
                RegistryOverlayImpl_GetEndpoints_Test.createCfgServiceMock(PETALS_REGISTRY_NODE.getGroupName(),
                        PETALS_REGISTRY_NODE.getPassword(), PETALS_REGISTRY_NODE.getHost(),
                        String.valueOf(PETALS_REGISTRY_NODE.getPort()), topologyName, REMOTE_CONTAINER_NAME));
        try {
            localRegistryOverlayImpl.start();
            remoteRegistryOverlayImpl.start();

            final Location localLocation = new Location(LOCAL_CONTAINER_NAME, "");

            final Location remoteLocation = new Location(REMOTE_CONTAINER_NAME, "");

            // Add the local internal endpoint
            localRegistryOverlayImpl.registerInternalEndpoint(LOCAL_INTERNAL_ENDPOINT_NAME, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", null);

            // Add the local external endpoint
            localRegistryOverlayImpl.registerExternalEndpoint(
                    new MockEndpoint(LOCAL_EXTERNAL_ENDPOINT_NAME, SERVICE_NAME_1, new QName[] { INTERFACE_NAME_1 }),
                    "");

            // Add the remote internal endpoint
            remoteRegistryOverlayImpl.registerInternalEndpoint(REMOTE_INTERNAL_ENDPOINT_NAME, SERVICE_NAME_1,
                    new QName[] { INTERFACE_NAME_1 }, "", null);

            // Add the remote external endpoint
            remoteRegistryOverlayImpl.registerExternalEndpoint(
                    new MockEndpoint(REMOTE_EXTERNAL_ENDPOINT_NAME, SERVICE_NAME_1, new QName[] { INTERFACE_NAME_1 }),
                    "");

            // Check that endpoints are registered
            final Collection<PetalsServiceEndpoint> registeredSvcEp = localRegistryOverlayImpl.getEndpoints(null);
            assertNotNull("Null list returned", registeredSvcEp);
            assertEquals("Invalid response number retrieving endpoints for service", 4, registeredSvcEp.size());
            boolean localInternalEpRegistered = false;
            boolean localExternalEpRegistered = false;
            boolean remoteInternalEpRegistered = false;
            boolean remoteExternalEpRegistered = false;
            for (final PetalsServiceEndpoint svcEp : registeredSvcEp) {
                if (svcEp.getEndpointName().equals(LOCAL_INTERNAL_ENDPOINT_NAME)) {
                    localInternalEpRegistered = true;
                    assertEquals(localLocation, svcEp.getLocation());
                } else if (svcEp.getEndpointName().equals(LOCAL_EXTERNAL_ENDPOINT_NAME)) {
                    localExternalEpRegistered = true;
                    assertEquals(localLocation, svcEp.getLocation());
                } else if (svcEp.getEndpointName().equals(REMOTE_INTERNAL_ENDPOINT_NAME)) {
                    remoteInternalEpRegistered = true;
                    assertEquals(remoteLocation, svcEp.getLocation());
                } else if (svcEp.getEndpointName().equals(REMOTE_EXTERNAL_ENDPOINT_NAME)) {
                    remoteExternalEpRegistered = true;
                    assertEquals(remoteLocation, svcEp.getLocation());
                } else {
                    fail("Not activeated endpoint returned");
                }
            }
            assertTrue("The local internal endpoint is missing", localInternalEpRegistered);
            assertTrue("The local external endpoint is missing", localExternalEpRegistered);
            assertTrue("The remote internal endpoint is missing", remoteInternalEpRegistered);
            assertTrue("The remote external endpoint is missing", remoteExternalEpRegistered);

            // get all internal endpoints
            final Collection<PetalsServiceEndpoint> allInternalEndpoints = localRegistryOverlayImpl
                    .getEndpoints(EndpointType.INTERNAL);
            assertNotNull("Null list returned", allInternalEndpoints);
            assertEquals("Unexpected number of internal endpoints", allInternalEndpoints.size(), 2);
            localInternalEpRegistered = false;
            localExternalEpRegistered = false;
            remoteInternalEpRegistered = false;
            remoteExternalEpRegistered = false;
            for (final PetalsServiceEndpoint endpoint : allInternalEndpoints) {
                if (endpoint.getEndpointName().equals(LOCAL_INTERNAL_ENDPOINT_NAME)) {
                    localInternalEpRegistered = true;
                } else if (endpoint.getEndpointName().equals(LOCAL_EXTERNAL_ENDPOINT_NAME)) {
                    localExternalEpRegistered = true;
                } else if (endpoint.getEndpointName().equals(REMOTE_INTERNAL_ENDPOINT_NAME)) {
                    remoteInternalEpRegistered = true;
                } else if (endpoint.getEndpointName().equals(REMOTE_EXTERNAL_ENDPOINT_NAME)) {
                    remoteExternalEpRegistered = true;
                } else {
                    fail("Not activated endpoint returned");
                }
            }
            assertTrue("The local internal endpoint is missing", localInternalEpRegistered);
            assertFalse("The local external endpoint is returned", localExternalEpRegistered);
            assertTrue("The remote internal endpoint is missing", remoteInternalEpRegistered);
            assertFalse("The remote external endpoint is returned", remoteExternalEpRegistered);

            // get all external endpoints
            final Collection<PetalsServiceEndpoint> allExternalEndpoints = localRegistryOverlayImpl
                    .getEndpoints(EndpointType.EXTERNAL);
            assertNotNull("Null list returned", allExternalEndpoints);
            assertEquals("Unexpected numer of external endpoints", allExternalEndpoints.size(), 2);
            localInternalEpRegistered = false;
            localExternalEpRegistered = false;
            remoteInternalEpRegistered = false;
            remoteExternalEpRegistered = false;
            for (final PetalsServiceEndpoint endpoint : allExternalEndpoints) {
                if (endpoint.getEndpointName().equals(LOCAL_INTERNAL_ENDPOINT_NAME)) {
                    localInternalEpRegistered = true;
                } else if (endpoint.getEndpointName().equals(LOCAL_EXTERNAL_ENDPOINT_NAME)) {
                    localExternalEpRegistered = true;
                } else if (endpoint.getEndpointName().equals(REMOTE_INTERNAL_ENDPOINT_NAME)) {
                    remoteInternalEpRegistered = true;
                } else if (endpoint.getEndpointName().equals(REMOTE_EXTERNAL_ENDPOINT_NAME)) {
                    remoteExternalEpRegistered = true;
                } else {
                    fail("Not activated endpoint returned");
                }
            }
            assertFalse("The local internal endpoint is returned", localInternalEpRegistered);
            assertTrue("The local external endpoint is missing", localExternalEpRegistered);
            assertFalse("The remote internal endpoint is returned", remoteInternalEpRegistered);
            assertTrue("The remote external endpoint is missing", remoteExternalEpRegistered);

        } finally {
            remoteRegistryOverlayImpl.stop();
            localRegistryOverlayImpl.stop();
        }
    }

}
